@file:JvmName("_AutocloseKt")

package ru.landgrafhomyak.db.serdha0.user_commons.executors

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.db.serdha0.api.LowLevelApi
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.OutputRow


@Suppress("FunctionName")
@PublishedApi
internal inline fun <R> _safeAutoClose(onAbort: () -> Unit = {}, onSuccess: () -> Unit = {}, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onAbort, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.EXACTLY_ONCE)
	}
	val ret: R
	var wasError = false
	try {
		ret = action()
	} catch (e1: Throwable) {
		wasError = true
		try {
			onAbort()
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	} finally {
		if (!wasError)
			onSuccess()
	}
	return ret
}

@Suppress("ClassName")
public class _ReturnFromInputError : Error("Returning from outer function from input row initializer block is undefined behaviour")

@Suppress("FunctionName", "REDUNDANT_INLINE_SUSPEND_FUNCTION_TYPE")
@LowLevelApi
public suspend inline fun <I : InputRow._Scope<*, O>, O : OutputRow._Iterator<*, Unit>> _safeAutoClose_IO(
	iRow: I,
	inputAction: suspend (I) -> Unit,
	outputAction: suspend (O) -> Unit,
) {
	contract {
		callsInPlace(inputAction, InvocationKind.EXACTLY_ONCE)
		callsInPlace(outputAction, InvocationKind.EXACTLY_ONCE)
	}
	var crossReturned = true
	_safeAutoClose(
		onAbort = { iRow._abort() },
		action = { inputAction(iRow); crossReturned = false },
		onSuccess = {
			if (crossReturned) throw _ReturnFromInputError()
			val oRow = iRow._finish()
			_safeAutoClose(
				onAbort = { oRow._abort() },
				action = { outputAction(oRow) },
				onSuccess = { oRow._finish() }
			)
		}
	)
}

@Suppress("FunctionName", "REDUNDANT_INLINE_SUSPEND_FUNCTION_TYPE")
@LowLevelApi
public suspend inline fun <I : InputRow._Scope<*, Unit>> _safeAutoClose_I(
	iRow: I,
	inputAction: suspend (I) -> Unit,
) {
	contract {
		callsInPlace(inputAction, InvocationKind.EXACTLY_ONCE)
	}
	_safeAutoClose(
		onAbort = { iRow._abort() },
		action = { inputAction(iRow) },
		onSuccess = { iRow._finish() }
	)
}

@Suppress("FunctionName", "REDUNDANT_INLINE_SUSPEND_FUNCTION_TYPE")
@LowLevelApi
public suspend inline fun <O : OutputRow._Iterator<*, Unit>> _safeAutoClose_O(
	oRow: O,
	outputAction: suspend (O) -> Unit,
) {
	contract {
		callsInPlace(outputAction, InvocationKind.EXACTLY_ONCE)
	}
	_safeAutoClose(
		onAbort = { oRow._abort() },
		action = { outputAction(oRow) },
		onSuccess = { oRow._finish() }
	)
}
public fun test() {

}