package ru.landgrafhomyak.db.serdha0.user_commons.types

import ru.landgrafhomyak.db.serdha0.api.misc.DatabaseType
import ru.landgrafhomyak.utility.IntSerializers

@OptIn(ExperimentalUnsignedTypes::class)
public abstract class STRING : DatabaseType._VirtualType<String, UByteArray, DatabaseType.BINARY_DATA> {
	public val maxLengthOrNull: UInt?
	public val isFixedSize: Boolean

	protected constructor(context: DatabaseType.Provider) : super(context.BINARY_DATA, allowValuesCaching = true) {
		this.maxLengthOrNull = null
		this.isFixedSize = false
	}

	protected constructor(wrappedType: DatabaseType.BINARY_DATA, maxLength: UInt, isFixedSize: Boolean) : super(wrappedType, allowValuesCaching = true) {
		this.maxLengthOrNull = maxLength
		this.isFixedSize = isFixedSize
	}

	override fun _equals(l: String, r: String): Boolean {
		return l == r
	}

	override fun _compare(l: String, r: String): Int {
		return l.compareTo(r)
	}

	public class ASCII : STRING {
		public constructor(context: DatabaseType.Provider) : super(context)

		public constructor(
			context: DatabaseType.Provider,
			maxLength: UInt,
			isFixedSize: Boolean
		) : super(context.BINARY_DATA(maxLength, isFixedSize), maxLength, isFixedSize)

		override val name: String = "STRING.ASCII"

		override fun _unwrap(w: UByteArray): String {
			if (this.maxLengthOrNull != null && (this.isFixedSize && w.size.toUInt() != this.maxLengthOrNull || w.size.toUInt() > this.maxLengthOrNull))
				throw RuntimeException("Encoded string has wrong size") // todo
			val out = CharArray(w.size)
			w.forEachIndexed { i, b ->
				if (b >= 128u)
					throw RuntimeException("Serialization error: char code greater than 127") // todo
				out[i] = Char(b.toUShort())
			}
			return out.concatToString()
		}

		override fun _wrap(w: String): UByteArray {
			if (this.maxLengthOrNull != null && (this.isFixedSize && w.length.toUInt() != this.maxLengthOrNull || w.length.toUInt() > this.maxLengthOrNull))
				throw RuntimeException("String has wrong size") // todo
			val out = UByteArray(w.length)
			w.forEachIndexed { i, c ->
				if (c.code >= 128)
					throw RuntimeException("Serialization error: char code greater than 127") // todo
				out[i] = c.code.toUByte()
			}
			return out
		}

	}

	public class UTF8 : STRING {
		public constructor(context: DatabaseType.Provider) : super(context)

		public constructor(
			context: DatabaseType.Provider,
			maxLength: UInt,
			isFixedSize: Boolean
		) : super(context.BINARY_DATA(maxLength * 4u, false), maxLength, isFixedSize)

		override val name: String = "STRING.UTF8"

		override fun _unwrap(w: UByteArray): String {
			if (this.maxLengthOrNull != null && this.isFixedSize && w.size.toUInt() != this.maxLengthOrNull)
				throw RuntimeException("Encoded string has wrong size") // todo
			val out = w.asByteArray().decodeToString()
			if (this.maxLengthOrNull != null && out.length.toUInt() > this.maxLengthOrNull)
				throw RuntimeException("Encoded string has wrong size") // todo
			return out
		}

		override fun _wrap(w: String): UByteArray {
			if (this.maxLengthOrNull != null && (this.isFixedSize && w.length.toUInt() != this.maxLengthOrNull || w.length.toUInt() > this.maxLengthOrNull))
				throw RuntimeException("String has wrong size") // todo
			return w.encodeToByteArray().asUByteArray()
		}
	}

	public class UTF16 : STRING {
		public constructor(context: DatabaseType.Provider) : super(context)

		public constructor(
			context: DatabaseType.Provider,
			maxLength: UInt,
			isFixedSize: Boolean
		) : super(context.BINARY_DATA(maxLength * 4u, false), maxLength, isFixedSize)

		override val name: String = "STRING.UTF16"

		override fun _unwrap(w: UByteArray): String {
			val length = w.size.toUInt() / 2u
			if (this.maxLengthOrNull != null && (this.isFixedSize && length != this.maxLengthOrNull || length > this.maxLengthOrNull))
				throw RuntimeException("Encoded string has wrong size") // todo
			return CharArray(length.toInt()) { i -> Char(IntSerializers.decode16beHu(w, i * 2)) }.concatToString()
		}

		override fun _wrap(w: String): UByteArray {
			if (this.maxLengthOrNull != null && (this.isFixedSize && w.length.toUInt() != this.maxLengthOrNull || w.length.toUInt() > this.maxLengthOrNull))
				throw RuntimeException("String has wrong size") // todo
			val out = UByteArray(w.length * 2)
			w.forEachIndexed { i, c ->
				IntSerializers.encode16beHu(out, i * 2, c.code.toUShort())
			}
			return out
		}
	}
}